/**
 * RNG and distribution objects.
 * @author J. Vozmediano
 * @date 2003 02
 * @bugs  jvt@trajano.us.es
 *
 */
#ifndef DISTRIB_H
#define DISTRIB_H

#include <math.h>

/**
 * RNG abstract class. Abstract class that contains an
 * implementation of the Wichmann-Hill composite linear
 * congruential random number generator.
 *
 * Must be implemented in derived classes. The derived GetVal(void)
 * will call Pick(void) and will operate with the result to give 
 * succesive random numbers properly distributed. 
 */
class Distrib
{
private:
  static unsigned long int seed_; //!< Seed
  static inline signed long int Trunc(double x) {
	return (x >=0) ? (int(floor(x))) : (int(ceil(x)));
  };
  
protected:      

  signed long int x_, y_, z_;
  /**
   *
   */
  Distrib (void);
  /**
   * Picks next random number.
   * Succesive calls return evenly distributed random numbers in [0,1)
   * It is available only to derived classes. 
   */
  double Pick(void);
  
public:
  virtual ~Distrib (void);

  /**
   * Gets next random number according to the class' distribution.
   *
   * @returns next random number generated
   * according to a distribution determined by the derived class.
   */
  virtual double  GetVal (void) = 0;
  /**
   * Seed initiator, must be called before any derived "GetVal".
   * If no seed is given or seed==0, a random one will be 
   * automatically chosen 
   *
   * @param seed 
   */
  static void InitRng (unsigned long int seed=0);
};


/**
 * Implements a constant distribution. Always returns the same value.
 */
class Constant : public Distrib {
private:
  double val_;
public:
  Constant(double value);
  /**
   * Gets next random number according to this distribution.
   */
  double GetVal(void);
};

/**
 * Implements a Uniform distribution in the interval [0,1)
 */
class Unif01 : public Distrib {
public:
  Unif01(void);
  /**
   * Gets next random number according to this distribution.
   */
  double GetVal(void);
};

/**
 * Implements a Uniform distribution in the interval [a,b)
 */
class Uniform : public Distrib {
private:
  double a_;
  double b_;
public:
  Uniform(double a, double b);
  /**
   * Gets next random number according to this distribution.
   */
  double GetVal(void);
};


/**
 * Implements a negative exponential distribution of mean m.
 */
class Exponential : public Distrib {
private:
  double mean_;
public:
  Exponential (double m);
  /**
   * Gets next random number according to this distribution.
   */
  double GetVal(void);
};

/**
 * Implements a normal distribution of mean m and
 * standard deviation sDev.
 */
class Normal : public Distrib {
private:
  double mean_;
  double sDev_;
public:
  Normal (double m, double sDev);
  /**
   * Gets next random number according to this distribution.
   */
  double GetVal(void);
};

/**
 * Implements an Erlang distribution of order k and mean m.
 */
class ErlangK: public Distrib {
private:
  double k_;
  double mean_;
public:
  ErlangK (unsigned int k, double m);
  /**
   * Gets next random number according to this distribution.
   */
  double GetVal(void);
};

/**
 * Implements a Pareto distribution of parameters scale and alpha.
 * Note that neither scale_ nor alpha_ are the mean value. 
 */
class Pareto: public Distrib {
private:
  double scale_;
  double alpha_;
public:
  Pareto (double scale, double alpha);
  /**
   * Gets next random number according to this distribution.
   */
  double GetVal(void);
};

/**
 * Implements a Hyperexponential distribution of mean m and
 * standard deviation sDev.
 */
class HyperExponential: public Distrib {
private:
  double mean_;
  double sDev_;
  double p_;
public:
  HyperExponential(double m, double sDev);
  /**
   * Gets next random number according to this distribution.
   */
  double GetVal(void);
};


/**
 * Implements a Discrete Empirical distribution.
 */
class DiscreteEmpirical: public Distrib {
 private:
  typedef struct {
	double Fx;
	double val;
  } valT_;
  valT_ *accumP_;
  int n_;
 public:
  /**
   * @param P[] probability function values 
   * @param V[] associated value corresponding to the probability value
   * @param n number of values
   */
  DiscreteEmpirical (const double P[],const double V[], int n);
  virtual ~DiscreteEmpirical (void);    
  double GetVal (void);
};

#endif

