/**
 * Discrete event simulator
 * @author J. Vozmediano
 * @date 2003 02
 * @bugs  jvt@trajano.us.es
 *
 */
#ifndef EVSIM
#define EVSIM

#include <queue>
using namespace std;

#include <string.h>
#include <stdlib.h>
#include "timeunits.h"

/**
 * Generic event
 *
 */
class Event 
{
private:
  timeunits evTime_; // Activation time
  unsigned long id_; // Event identifier. Use as desired.
  char     name_[6]; // Event name. Use as desired.

public:
  /**
   * Simple constructor. Creates an anonymous but potentially
   * identifiable event, to be fired at time t.  
   * @param t time to fire the event 
   * @param id event identifier
   */
  Event(timeunits t, unsigned long id=0):
	evTime_(t), id_(id) { };

  /**
   * Complete constructor. Creates an event.
   * @param t time to fire the event 
   * @param name name to identify the event
   * @param id event numeric identifier
   */
  Event(timeunits t, char* name, unsigned long id = 0):
	evTime_(t), id_(id) 
	{
	  if (0 != name){
		for (unsigned int i=0; (i<5)&&(i<strlen(name));i++){
		  name_[i] = *(name+i);
		}
		name[5]=0;
	  };
	}

  /**
   * Destructor. MUST BE virtual, to free the name string.
   */
  virtual ~Event(void) {
  };

  /**
   * Event-handler function, to be called at activation time
   * Being event-type dependent, it must be defined in derived classes
   */
  virtual void     Handle(void) = 0;

  /**
   * @returns activation time
   */
  inline timeunits Time(void) const {return evTime_;};
  /**
   * Modifies activation time
   * No check is made to avoid setting past times.
   * @param t time to fire the event
   */
  inline void      SetTime(timeunits t) {evTime_ = t;};
  /** 
   * @return event's name
   */
  char*            GetName(void) {return name_;};
  /** 
   * @return event's numerical identifier
   */
  inline unsigned int GetId(void) const {return id_;};
};


/**
 * Compares two events' activation time
 */
class CmpEvents
{
public:
/**
 * Compares two events' activation time
 * @return true it activation time of p1 is strictly greater than of p2
 */
  inline bool operator() (Event*& p1, Event*& p2) const{
    return p1->Time() > p2->Time();
  }
};


/**
 * Event simulator container.
 */
class EvSim
{
private:
  static double version_;
  timeunits now_;       //Current time
  priority_queue <Event*,vector<Event*>, CmpEvents> evQ_;
  timeunits from_;      //Simulation Starting Time
  timeunits to_;        //Simulation Ending Time
  unsigned int len_;    //Event queue length
  unsigned long evnum_; //Event counter

public:
  /**
   * Constructor. Creates an event simulator object
   */
  EvSim(void): now_(0), from_(0), to_(MAXTIME),len_(0), evnum_(0) {};
  /**
   * Destroys the object and dumps some simulator statistics.
   */
  virtual ~EvSim(void); 
  /**
   * Get current simulated time
   * @return current time in timeunits (seconds, tics...)
   */
  inline timeunits CurrentTime(void) const {return now_;};
  /**
   * Start the simulator scheduler and start processing events from initial time t final time.
   * @param from initial time, defaults to 0
   * @param to   final time, defaults to MAXTIME
   */
  void Run(timeunits from = 0, timeunits to = MAXTIME);
  /**
   * Enqueue an event to occur in the future. If more than one
   * event has to occur at the same time, enqueue it either in FIFO
   * order or in LIFO order.
   * @param e pointer to the event to enqueue
   * @param fifo enqueueing tiebreaker
   */
  void Schedule(Event* e, bool fifo = true) ;

  /**
   * Dumps simulation-related statistics. These include number of
   * events, time of the last event, event rate... This function is
   * automatically called by the destructor, so you do not need to
   * call it directly.
   */
  void dumpStats (void);
  /**
   * Chech library version, abort if it does not match with the internal version number.
   * @param to check
   */
  static void AssertVersion (double version);
};

/**
 * Unique EvSim instance
 */
extern EvSim Simulation;

#endif

