/**
 * Statistical gatherers
 * @author J. Vozmediano
 * @date 2003 02
 * @bugs  jvt@trajano.us.es
 *
 */
#ifndef LSTAT_H
#define LSTAT_H

#include <list>
/**
 * Included for time-dependent statistics
 */
#include "ev.h"

/**
 * Time-insensitive statistics
 */
class LStat
{
private:
  static  list<LStat*> Statistics_;
  bool    fulldump_;
  double  n_;         //!< Number of samples
  double  maximum_;   //!< Maximum sample
  double  minimum_;   //!< Mimimum sample
  double  m_;         //!< Average estimator
  double  variance_;  //!< Variance estimator
  const char*   name_;      //!< Descriptive name for the statistical gatherer

protected: //Available only to derived classes.
  inline void SetN(double x)   {n_=x;};
  inline void SetMax(double x) {maximum_=x;};
  inline void SetMin(double x) {minimum_=x;};
  inline void SetVar(double x) {variance_=x;};
  inline void SetMean(double x){m_=x;};
  void        SetFullDump (bool d);
  
public:
  /**
   * Constructor.
   * Creates a statitical gatherer
   * @param n descriptive name
   */
  LStat (const char* n=0);
  /**
   * Destructor.
   * Automatically dumps statistical data
   */
  virtual ~LStat (void);

  /**
   * Stores a new value in the statistical object
   * @param x value to account for
   */
  virtual void accum (double x); 
  /**
   * @returns number of samples collected
   */
  double num (void);
  /**
   * @returns maximum sample value
   */
  double maxval (void);
  /**
   * @returns minimum sample value
   */
  double minval (void);
  /**
   * @returns the unbiased average estimator
   */
  double mean (void);
  /**
   * @returns the unbiased variance estimator
   */
  double var (void);
  /**
   * @returns  unbiased standard deviation estimator
   */
  double sDev (void);
  /**
   * @returns  mean root square
   */
  double rMeanSq (void);
  /**
   *
   *
   */
  void dumpStats (void);
  /**
   * Reset this statistical gatherer.
   */
  virtual void   reset(void);
  /**
   * Reset all statistical gatherers.
   */
  static void   resetAll(void);
};

/**
 * Time-averaged statistics
 */
class TStat : public LStat 
{
private:
  timeunits prevt_;  //Time instant of last sample
  double    prevv_;  //Value of last sample.
  EvSim* Sim_;       //Pointer to the simulator object, for 
  //clock recovery purposes

public:
  /**
   * Constructor.
   * Creates a timed statitical gatherer
   * @param n descriptive name
   * @param Sim pointer to the simulator object to get time information.
   */
  TStat (const char* n, EvSim* Sim);
  /**
   * Destructor.
   * Automatically dumps statistical data
   */
  ~TStat(void);
  /**
   * Stores a new value in the statistical object
   * @param x value to account for
   */
  virtual void   accum (double x);
  /**
   * Reset this statistical gatherer.
   */
  virtual void   reset(void);
};
#endif

