/**
 * Elementary network components
 * @author J. Vozmediano
 * @date 2003 02
 * @bugs  jvt@trajano.us.es
 *
 */
#ifndef NETELEM_H
#define NETELEM_H

#include <queue>
#include <map>
#include <vector>
#include <list>
#include <stdio.h>

using namespace std;

#include "distrib.h"
#include "lstat.h"
#include "ev.h"

class Packet;
class NetElement;
class PacketSource;
class PacketSink;
class Server;
class SrcSink;
class Fragmenter;


/**
 * Abstract class NetElement, to be derived onto Sources, 
 * Sinks and Servers. 
 *
 * Contains member functions to attach one NetElement to others
 * Possible arrangements are 1 to 1, 1 to N and N to 1.
 * The attachment enables this NetElement to pass 
 * packets to the attached ones. 
 *
 */
class NetElement 
{
public:
  /**
   * @return unique element identifier.
   */
  unsigned int              GetId(void) const;

  /**
   * @param k donwstream NetElement to retrieve
   * @return pointer to the k-th NetElement this one is 
   * attached to.
   */
  NetElement*      GetTo(unsigned int k) const;

  /**
   * @return number of elements this one is connected to.
   */
  inline unsigned int GetSizeTo(void) const {
    return tovector_.size();
  };

  /**
   * Sends a packet to the downstream-attached NetElements.
   * Depending on the packet's data, sending may be either unicast
   * (to one adjacent NetElement, as indicated by the packet) or
   * broadcast (to every adjacent NetElement, if no information
   * is provided by the packet). In the latter case, packets are
   * replicated as needed. This member function can be overrided
   * (but this is usually unnecessary). Defaults to calling the
   * Recv member function of the downstream-attached NetElement
   *
   * @param now current time
   * @param p pointer to packet to be sent 
   */
  virtual void     Send     (timeunits now, Packet* p);

  /**
   * Receives a packet from one of the upstream NetElement attached to this one.
   * To be implemented in derived classes.
   *
   * @param now current time
   * @param p   pointer to packet to be sent
   */
  virtual void     Recv     (timeunits now, Packet* p) = 0;  

  /**
   * To be implemented in every derived class, this function resets
   * every statistical measure and counter of this NetElement. This
   * eliminates the transient period effect. Must be kept public in
   * the derived classes.
   */
  virtual void     Reset     (bool dump=false) = 0;  
  /**
   * Connects the output of this NetElement to the argument's
   * input. The pointed element address is stored in the last
   * position of a STL vector. 
   *
   * @param to pointer to the downstream NetElement to attach to.
   * @return a pointer to the next NetElement.
   */
  NetElement*      ConnectTo(NetElement* to);  

  /**
   * FragmentDrop will be chain-called by attached NetElements
   * up to the last one, to immediately inform of a packet drop. 
   * This is useful to keep track of packet or fragment statistics.
   * 
   * Some NetElements (typically Sinks) may not implement
   * FragmentDrop, so this base-class member function is provided
   * as last resort. It will copy the packet and call the
   * FragmentDrop member function of every downstream-attached NetElement
   *
   * @param p pointer to the dropped packet. Route information can be 
   *          retrieved from the packet, allowing calling only the 
   *          neccesary NetElements 
   */
  virtual void     FragmentDrop (Packet* p);

  virtual ~NetElement(void);    
  /**
   * Activates the traces for all NetElements. Has to be called 
   * before creating any NetElement.
   * The trace file has nam-compatible format.
   */
  static void traceOn(char* filename =0);
  /**
   * For trace use only.
   * @return true  if tracing is activated.
   * @return false otherwise
   */
  bool IsTraceing(void) const {return namtraceflag;};
  /**
   * For trace use only.
   * @return the trace file.
   */
  FILE* traceFile(void) const {return fo;}
  /**
   * Logs a trace line in the trace file.
   * 
   * @param t time at which the event takes place
   * @param event  nam-compatible event indicator. 
   *               Can be either '+', 'd', '-' or 'h'.
   * @param p      pointer to the packet to log.
   *               Other relevant data are taken from the
   *               packet.
   */
  void packetTrace(timeunits t, char event, Packet* p);
  /**
   * Enqueues a special event to reset the statistical measures of
   * every NetElement at a given time. This includes not only
   * statistical acumulators, but also any private counter the
   * specific NetElements may have.
   * 
   * @param t time at which the statistics should be reseted. 
   */
  static void Transit (timeunits t);

  /**
   *
   *
   */
  static void Batch (timeunits tinitial, timeunits lapse, int batches);
  /**
   * Runs through an STL list that contains pointers to every
   * constructed NetElement and invoke its Reset() member function.
   * This static member function is automatically used by Transit.
   */
  static void resetAll(bool dump=false);

private:
  //Trace-related variables
  static bool namtraceflag;
  static FILE* fo;

  static unsigned int numElements_;
  static list<NetElement*> resetList_;

  unsigned int     id_;
  vector <NetElement*> tovector_;  //!< Stores the outputs of the Netelement

protected:

  NetElement (void);
  virtual void     namConnectTrace (void);
};

/**
 * Implements a generic packet or fragment. Packets are generated in a PacketSource, 
 * can be stored in Server Queues, and are destroyed in Sinks.
 */
class Packet {
private:
  static           unsigned long int totalPackets_;
  unsigned long int id_;          //!< Identifier (running counter)
  NetElement*      source_;       //!< Pointer to source 
  NetElement*      destination_;  //!< Pointer to sink
  unsigned int     bits_;         //!< Packet/fragment size in bits
  timeunits        timestamp_;    //!< Time at which the original packet was born
  unsigned long    datagramid_;  //!< identifies a non fragmented Packet
  unsigned int     totalbits_;   //!< Original Packet size in bits
  unsigned int     currenthop_;  //!< Number of performed hops
  
public:
  /**
   * Constructor.
   * @param src pointer to the source NetElement, typically a PacketSource
   * @param dst pointer to the destination NetElement, typically a PacketSink
   * @param bits packet size in bits. Has to be multiple of 8.
   * @param stamp generation time of the packet.
   * @param datagramid identifies the original packet a fragment packet belongs to.
   * @param totalbits datagram size in bits. Has to be multiple of 8.
   * @param k puts the packet in the k-th hop alogn the route from src to dst.
   */
  Packet (NetElement* src, NetElement* dst, 
          unsigned int bits, timeunits stamp, 
          unsigned long datagramid = 0, 
          unsigned int totalbits = 0, 
          unsigned int k = 0);
  virtual ~Packet(void);
  /**
   * @return the origination time of the packet. 
   */
  inline timeunits GetStamp(void) const {return timestamp_;};
  /**
   * @return the packet or fragment unique identifier.
   */
  unsigned long int         GetPacketId(void) const;
  /**
   * @return the size, in bits, of this packet or fragment.
   */
  unsigned int              GetBits (void)    const;
  /**
   * @return the size, in bits, of the original packet, before any fragmentation.
   */
  unsigned int     GetTotalBits (void) const;
  /**
   * @return the original packet identifier.
   */
  unsigned long    GetDatagramId (void) const;
  /**
   * @return pointer to the originating NetElement
   */
  NetElement*      GetSrc(void)       const;
  /**
   * @return pointer to the destination NetElement
   */
  NetElement*      GetDst(void)       const;
  /**
   * @return current position of this packet along the route from
   * src to dst, in number of hops
   */
  unsigned int     GetPos (void) const;
  /**
   * Advances the packet to the next hop to dst.
   */
  void             Jump(void);
  /**
   * Modifies the original packet length
   * @param b packet size in bits.
   */
  void             SetTotalBits (unsigned int b);
  /**
   * Modifies the original packet length
   * @param B packet size in octets.
   */
  void             SetBytes (unsigned int B);
  /**
   * Modifies this fragment size.
   * @param b fragment size in bits.
   */
  void             SetBits (unsigned int b);
  /**
   * @param id Changes the unique datagram identifier.
   */
  void             SetDatagramId (unsigned long id);
  /**
   * Modifies the origination time
   * @param stamp time to stamp in the Packet.
   */
  void             Stamp(timeunits stamp); 
};

/**
 * Timer event. Each Timer is associated to a Packet, to a
 * NetElement that is expected to receive this packet when the
 * Timer expires, and to a handling member function in the
 * NetElement that will handle the Packet.
 */
class Timer : public Event {
public:
  /**
   * Handling function, takes an instant, a (pointer to a) packet
   * and returns a (pointer to a) NetElement.
   */
  typedef void (NetElement::* TimerHandler)(timeunits, Packet*);

  /**
   * Creates a timed event.
   * @param t   time instant to fire the event 
   * @param id  unique event identifier
   * @param to  pointer to the NetElement which 
   *            is expected to receive the packet.
   * @param th  handling function to be called, typically NetElement::Recv().
   * @param p   pointer to the packet to be handled.
   */
  Timer(timeunits t, unsigned int id, 
        NetElement* to, TimerHandler th, 
        Packet* p);
  virtual ~Timer(void);
private:
  void              Handle(void);
  NetElement*       to_;
  TimerHandler      OnExpiry_; //to be called when this timer expires
  Packet*           packet_;   //packet this timer refers to
};
/**
 * Packet source generates packets according to a given distribution
 * with sizes following a (second) given distribution
 */
class PacketSource : public NetElement {
private:
  Distrib*          interarrival_; //Burst Inter-arrival time, in timeunits
  Distrib*          size_;         //Packet size, in bits
  LStat             S_interArrival; //Statistics
  vector <NetElement*> route_;
  LStat S_packetDelay_;
  //  virtual void      namConnectTrace (void);
  /**
   * Enqueues next generation instant and sends the packet to the
   * next NetElement, calling NetElement::Send.
   * @param now time at which the sending action takes place
   * @place p   pointer to the packet to be sent.
   */
  virtual void      Send  (timeunits now,  Packet *p);
  /**
   * Empty. A source receives nothing.
   */
  void              Recv  (timeunits now,  Packet *p);

protected:
  unsigned long int numPackets_;

public:
  /**
   * Creates a PacketSource. 
   * @param interarrival pointer to a distribution that 
   *        describes the interarrival process between packets
   *        (in timeunits)
   * @param service pointer to a distribution that describes 
   *        the packet sizes (in bits)
   */
  PacketSource(Distrib* interarrival, Distrib* service);
  virtual ~PacketSource(void);
  /**
   * Starts the source
   * @param now time to start generating packets.
   */
  virtual void      Start (timeunits now);
  /**
   * Resets interarrival and packet delay statistics.
   */
  void              Reset (bool dump=false);
  /**
   * Access to the packet delay statistical collector. This function is called by 
   * the cooresponding SrcSink NetElement, if any.
   * @param t value to account for.
   */
  void              AccumDelay (timeunits t);
  /**
   * Succesive calls allow to build-up routes from the source
   * NetElement to the destination NetElement.  
   * @param ne pointer to the NetElement to append 
   *           to the end of the current route.
   * @return this.
   */
  PacketSource*     AddHop (NetElement* ne);
  /**
   * Allows to get the a given NetElement in the route.
   * @param index ordinal of the NetElement to retrieve.
   * @return pointer to the next NetElement, or 0 if
   * at the end of the route.
   */
  inline NetElement* GetNextHop (unsigned int index) const {
    if(index<route_.size()) {
      return route_[index];
    }
    else {
      return 0;
    }
  }
};
/**
 * Simple Sink. Silently deletes received packets.
 */
class PacketSink : public NetElement {
private:
  /**
   * Empty.
   */
  void              Send  (timeunits now,  Packet *p);
public:
  PacketSink(void);  
  virtual ~PacketSink(void);
  /**
   * Silently destroy the packet. No statistics are collected.
   */
  virtual void      Recv  (timeunits now,  Packet *p);
  /**
   * Empty, kept for derived classes
   */
  void              Reset (bool dump=false){};
  void              namConnectTrace(void);
};
/**
 * Source sink. More sophisticated sink, keeps 
 * per-source count of received packets and 
 * fragments, as well as delay statistics.
 */
class SrcSink : public PacketSink {
private:
  map <unsigned int, unsigned long> srcmap;//source, count
  map <unsigned int, unsigned long> srcfragmentsmap;//source, count
  map <unsigned long, unsigned int> srcfragmentscount;//datagram,bits
  map <unsigned long, unsigned int> srcdroppedscount;//datagram,bits

public:
  SrcSink (void);
  virtual ~SrcSink (void);
  /**
   * Keeps count of number of received fragments and complete packets,
   * then delete the packet.
   * @param now current time, needed to compute the packet's delay.
   * @param p   pointer to the packet to be deleted.
   */
  virtual void Recv (timeunits now, Packet *p);
  /**
   * Similar to SrcSink::Recv(void), needs to be called for dropped packets.
   * This enables keeping count of lost framents, and avoids 
   * waiting to complete for a packet with a lost fragment.
   * per-source count of received packets and fragments.
   * @param p   pointer to the packet to be dropped.
   */
  virtual void FragmentDrop (Packet* p);
  /**
   * Resets counters of complete and incomplete packets and fragments.
   */
  void              Reset (bool dump=false);
};

/**
 * Supported queueing disciplines for the server
 */
enum queuediscipline {
  fifo, /**< first-in first-out */
  lifo, /**< last-in first-out */
  sjf   /**< shortest job first (smaller packet) */
};
/**
 * Implements the queueing discipline for the Server class. 
 */
class CmpPacket {
public:
  /**
   * Constructor.  
   * @param discipline queueing discipline. Defaults
   * to fifo. Other values are lifo and sjf.
   */
  CmpPacket (queuediscipline discipline = fifo) : discipline_(discipline) {
    // Empty
  }

  /**
   * Comparator. 
   * @param p1 pointer to the first object reference to compare
   * @param p2 pointer to the second object reference to compare
   * @return true if p1 must be served before
   * @return false elsewhere
   */
  inline bool operator() (Packet*& p1, Packet*& p2) const {
    // The return funtion is defined by the discipline used in the constructor
    switch (discipline_){
    case sjf: return p1->GetBits() > p2->GetBits();
    case lifo: return p1->GetStamp() < p2->GetStamp();
    case fifo:
    default: return p1->GetStamp() > p2->GetStamp();
    }
  }

private:
  queuediscipline discipline_; // Keeps the comparison discipline
};
/**
 * Models a generic server or multiplexer connected to a transmission link.
 */
class Server : public NetElement {
public:
  /**
   * Constructor.
   * @param name identifies the server in the statistics dump.
   * @param rBin server capacity in bits/timeunit
   * @param qsize queue size in bits, 0 means no queue and  -1 (default) means infinite.
   * @param delay additional propagation delay
   * @param discipline queueing discipline of the server
   * @param mtu maximum transfer unit. If the packet size is 
   * bigger, fragmentation will occur.
   * @param pebit transmission bit error probability
   */
  Server  (const char* name,           //Identifies the server
           double rBin,          //Server capacity in bits/timeunit
           int qsize=-1,         //Queue size, -1 means infinite
           timeunits delay = 0,  //Additional constant delay 
           //(e.g. propagation delay)
           queuediscipline discipline = fifo,  //Queueing discipline
           unsigned int mtu = 0, // Maximun Tranfer Unit 
           double pebit = 0); // Bit error probability
  /**
   * Dumps statistics on service time, time-averaged queue length,
   * observed queue length, server occupancy, and packet loss
   * ratio.
   */
  virtual ~Server(void);
  /** 
   * Passes the packet to the next NetElement or NetElement.
   * @param now  current time
   * @param p    pointer to the packet to be received
   */
  void          Send (timeunits now, Packet *p);
  /**
   * Receives a packet from the upstream NetElement
   * and either puts it in the server queue or starts 
   * transmiting it.
   * @param now  current time
   * @param p    pointer to the packet to be received
   */
  void          Recv (timeunits now, Packet *p);

  /**
   * Resets server's statiscs.
   */
  void          Reset (bool dump=false);

private:

  const char*        name_;
  double       rBin_; //In bits per timeunit
  int maxQ_;   //Max Queue Size, in bits, signed.
  timeunits   delay_; //In timeunits
  unsigned int mtu_; // In bits
  int inSystemPackets_; /* 0  when empty system
                         * 1  when queue empty and 
                         *     packet being served and
                         * >1 when packet being served and
                         *     others waiting
                         * signed, to compare to maxQ_ w/o warnings
                         */

  unsigned long int droppedPackets_;
  double pebit_; 
  
  Unif01 txnError_;

  int inSystemBits_; //output token bucket

  //Packet queue. Packets are sorted according to their timestamp_
  priority_queue <Packet*, vector<Packet*>, CmpPacket>* qPacket_; 
 
  
  LStat         S_serviceTime; //Service Time
  LStat         S_obsNumPackets;//Observed num of pkts in the server
  TStat         T_numPackets;   //Expected num of pkts in the server
  TStat         T_load;        //Server load

  void          DeQueue (timeunits now, Packet *p);
  void          namConnectTrace (void);
};

/**
 * Fragmenter object.
 * 
 */
class Fragmenter : public NetElement {
private:
  unsigned int mtu_; // Maximum Transfer Unit.
  void       namConnectTrace (void);
  void       Send  (timeunits now,  Packet *p);
  void       Recv  (timeunits now,  Packet *p);
public:
  /**
   * Creates a NetElement that receives packets and splits them in
   * blocks up to a Maximum Tranfer Unit. Note that it does not add
   * any "space" between fragments, i.e. the resulting fragments
   * will be sent "at once" with the same timestamp than the
   * originating packet.
   *  @param mtu maximum tranfer unit, in
   * bits. 0 if unlimited.
   */
  Fragmenter (unsigned int mtu = 0);
  /**
   * Empty (no counter to reset)
   */
  void              Reset (bool dump=false){};

  virtual ~Fragmenter(void);
};



/**
 * Auxiliary class to implement state-based sources, such as ON-OFF
 * sources.  A state lasts for a given duration, generating data
 * units of samplesize bits at a given rate.
 */
class SourceState {
private:
  Distrib* duration_;       // Duration, in timeunits
  Distrib* interarrival_;   // Trx. rate in this state, in bits/timeunit
  int      samplesize_;     // In bits
  int      sampleCount_;
public:
  /**
   * The distribution objects will be deleted by the destructor.
   * @param duration pointer to a distribution that 
   *        describes the permanence time in this state.
   * @param interarrival pointer to a distribution that 
   *        describes the interarrival process between packets.
   * @param samplesize constant size, in bits, of the packets 
   *        generated in this state. 
   * @note  it generates and enqueues a complete burst, to avoid wasting time.
   *        Hence, the state duration must be finite. If this is not the case, 
   *        don't use this NetElement.
   */
  SourceState(Distrib* duration, Distrib* interarrival, int samplesize);
  /**
   * Deletes the distribution objects.
   */
  virtual ~SourceState(void);
  /**
   * @return constant size, in bits, of the packets generated in this state.
   */
  int GetSampleSize(void) const;
  /**
   * @return next permanence time, in timeunits, in this state, according to the duration distribution.
   */
  timeunits GetDuration(void) const;
  /**
   * @return time to wait to generate the next packet, in timeunits, according to the interarrival distribution.
   */
  virtual timeunits GetInterarrival(void) const;
  /**
   *
   */
  inline void AddToPacketCount (int n) { sampleCount_ += n;}
};


/**
 * Two-state source.
 * @note to speed up the simulation
 * the full burst generated at any state is enqueued at 
 * once in the event list. 
 */
class BiStateSource : public PacketSource {
private:
  int state_;
  timeunits maxDelay_;
  SourceState *s_[2];
  int evid_;

  void              Reset (bool dump=false){};
  void              Send(timeunits now,  Packet *p);
public:
  /**
   * Creates a two-state alternating source.
   * @param on pointer to first (initial) SourceState
   * @param off pointer to second SourceState
   * @param maxdelay a generated packet 
   * can wait for the generation of more packets to be packed
   * together in the same transmission PDU up to maxdelay
   * timeunits. This parameter determines the maximum number of
   * packets that will be packed together in a PDU. Defaults to 0, 
   * meaning no waiting time (or one packet per PDU).
   */
  BiStateSource (SourceState* on, SourceState* off, timeunits maxdelay = 0);

  /**
   * Deletes the SourceState objects
   */
  virtual ~BiStateSource(void);
  /**
   * Activate the BiStateSource into its first SourceState
   * @param now time to fire the generation process
   */
  void Start(timeunits now);
};


/**
 * Adds headers and tails to the transmitted packets, roughly
 * modelling the overhead of the lower layer protocols.  
 */
class Enveloper : public NetElement {
private:
  unsigned int bitOverhead_;
  void       Send  (timeunits now,  Packet *p);
  void       Recv  (timeunits now,  Packet *p);
  void              Reset (bool dump=false){};
public:
  /**
   * Creates an envelope in which the data is wrapped before
   * transmission.
   * @param bOh bit overhead, i.e. number of overhead bits that
   * increase packet size. It should be an integer number of octets
   */
  Enveloper (unsigned int bOh);
  ~Enveloper (void){};
};


/**
 * Adds transmission overhead corresponding to RTP/UDP/IP header
 */
class VoIPEnveloper : public Enveloper {
private:
  void              Reset (bool dump=false){};
public:
  /**
   * Creates an envelope that adds transmission overhead corresponding to RTP/UDP/IP header
   */  
  VoIPEnveloper (void) : Enveloper(320){};
  ~VoIPEnveloper (void){};
};



/**
 * Voice source that takes samples from a file.
 */
class VoiceTraceFileSource : public PacketSource {
private:
  static const int maxEvents_;
  int fd_;
  bool loop_;
  int offset_;
  int nfpp_;
  timeunits tuPerSample_;
  char activeCode_;
  int  activeBitSize_;
  char sidCode_;
  int  sidBitSize_;
  int  position_;
  void Reset(bool dump=false){};
  void Recv(timeunits now, Packet *p){};
  void Send(timeunits now,  Packet *p);
public:
  /**
   * Creates a voice source that obtains real samples from a file.
   * @param infile name of the file containing sample frame types. The file format is a raw sequence of char identifying frame types (no spaces between frame identifiers)
   * @param loop play the same file continuously
   * @param offset position to start reading from
   * @param nfpp number of codec frames per packet
   * @param tuPerSample separation between two consecutive codec frames, in timeunits
   * @param activeCode char value that identifies active frames in the file
   * @param activeBitSize size of active frames, in bits
   * @param sidCode char value that identifies SID frames in the file
   * @param sidBitSize size of SID frames, in bits
   */
  VoiceTraceFileSource (const char* infile, 
                        bool loop,
                        int offset, 
                        int nfpp, 
                        timeunits tuPerSample,
                        char activeCode, int activeBitSize, 
                        char sidCode, int sidBitSize);
    
  /**
   * Activate the VoiceTraceFileSource
   * @param now time to start reading packets
   */
  void Start(timeunits now);
  virtual ~VoiceTraceFileSource (void);
};


#endif

